#!/usr/bin/perl 
#
# Czech Morphology Analyzer
# (C) 2000 Jan Hajic
#
# Parameters:
# 1 Dictionary name (complete dictionary: roots, endings)
# < Input file (iso-8859-2, clean-text/SGML/HTML file)
# > Output file (word per line, SGML (csts) format; see SetOptions for more)
#

use FileHandle; # for server mode (autoflush)
use Socket; # for server mode
use Carp; # for server mode

# Version INFO:
$FMVERSION = "1.1.0"; # unaccented mode handling added (i/o dict. dict.)
$FMDATE = "2001/08/22";
#$FMVERSION = "1.0.4"; # windows code handling added
#$FMDATE = "2001/08/16";
#$FMVERSION = "1.0.3"; # dict/code info passing from dictionary to client
#$FMDATE = "2000/08/17";
#$FMVERSION = "1.0.2"; # added (finished) correct handling of CSTS input
#$FMDATE = "2000/08/14";
#$FMVERSION = "1.0.1"; # correct comment (_^(...)) and _:T/W output; no &semi;
#$FMDATE = "2000/08/14";
#$FMVERSION = "1.0.0";
#$FMDATE = "2000/08/13";

# PARAMETERS/OPTIONS: see sub SetOptions, make a copy and set it there

# GLOBAL variables:

# USER-SUPPLIED OR DEFAULT OPTIONS/PARAMETERS/ARGUMENTS:
# see sub SetOptiions

int $fServ; # 0 (default): stdin/stdout or $1/$2 normal run; 
            # 1 - daemon run: server for client requests; set in main
$szSERVER = "SERVER";
my $iPort;
$szNOSTDIO = "NOSTDIO";
$szMAIn = ""; # input file name if not stdin
$szMAOut = ""; # output file name if not stdout

my $cEndingCount = 0; # just stats - ending entries
my $cRootCount   = 0; # just stats - root entries
my $szDictIDAndVersionInfo;

int $fInputType; # input file type (SGML, plain)
int $fSGMLType; # if input is SGML, type (HTML, other SGML, CSTS)
int $fSGML;  $fSGML =  1; # input file type
int $fCSTS;  $fCSTS =  4; # input file type
int $fHTML;  $fHTML =  2; # input file type
int $fPlain; $fPlain = 3; # input file type

int $ISOLATIN2; $ISOLATIN2 = 1;
int $WINCP1250; $WINCP1250 = 2;
int $fInputCode; # current input coding: 
                 # only ISO Latin2 (1) and Win CP 1250 supported (2)

$szPar1 = ""; # optional command-line parameter (after dictionary name, $2)
$szPar2 = ""; # second optional command-line parameter ($3) (unused)

my $fStripAccents; # 1 if accents should be stripped off dictionary

int $cEndingCountMax; # maximum number of characters in an ending in dictionary
                 # (set in GetDictionary)
# capitalization handling constants:
$iCapOrig = 0;
$iCapOnly = 1;
$iCapLower = 2;

# for case conversion, letter identification (GetToken and calls): ISO code
$szUpperAccented = "ݮϫǡ";
$szLowerAccented = "籶";

$reUpperAcc = "[$szUpperAccented]";
$reLowerAcc = "[$szLowerAccented]";

$opEnt = "&lt;";
$clEnt = "&gt;";
$entAmp = "&amp;";
$entSemi = "&semi;";
$entPercnt = "&percnt;";
$entSlash = "&slash;";
$entVbar = "&vbar;";

int $iRCTokenOK; 
    $iRCTokenOK = 1; # constant for next-token-available flag after GetToken

$STRICT = 1; # first pass when even conditional filters are in effect
$RELAXED = 2; # second pass if first produces nothing; conditional filters out

int $iPass; # for pass indication (STRICT, RELAXED)
$szFilterT = ""; # current tag filter, to be set depending on $iPass
$szFilterL = ""; # current (synt/sem/style) filter, to be set dep. on $iPass

# $hResultsCollected; # hash for results of a single wordform

sub StripAccents($) # strip accents from input string ##################
{
  my ($szIn) = @_;
  my $szOut;
  ($szOut = $szIn) =~ tr 
    [ݮϫǡ籶]
    [AEEIOUUYZSCRDTNLRLOOAUCASZLSZAACEEIDNOUTaeeiouuyzscrdtnlrlooaucaszlszaaceeidnout] ;
  return $szOut;
} # of StripAccents ####################################################

sub wee2il2($) # convert string from Win Cp1250 to ISO Latin 2 #########
{
  my ($szIn) = @_;
  my $szOut;
  ($szOut = $szIn) =~ tr 
        [\232\212\235\215\236\216\234\214\237\217\204\224\205] 
        ["".] ;
  return $szOut;
} # of sub wee2il2 ######################################################

sub il22wee($) # convert string from ISO Latin 2 to Win Cp1250 ##########
{
  my ($szIn) = @_;
  my $szOut;
  ($szOut = $szIn) =~ tr 
        []
        [\232\212\235\215\236\216\234\214\237\217] ;
  return $szOut;
} # of sub il22wee #######################################################

sub PrintCoded($) ### print coded string in win CP 1250 if input was same
{
  my ($szIn) = @_;

  if ($fInputCode == $WINCP1250) {
    print il22wee($szIn);
  }
  else {
    print $szIn;
  }
  return 0;
} # of sub PrintCoded ####################################################

sub Ch2Entity($) # code parameter (single char) to SGML entity ###########
{
  # returns input char if conversion not needed
  my ($chIn) = @_;
  my $szEntity;
  if ($chIn eq "<") {
    $szEntity = $opEnt;
  }
  elsif ($chIn eq ">") {
    $szEntity = $clEnt;
  }
  elsif ($chIn eq "&") {
    $szEntity = $entAmp;
  }
# semicolon, after all, does not make problems
#  elsif ($chIn eq ";") {
#    $szEntity = $entSemi;
#  }
  elsif ($chIn eq "%") {
    $szEntity = $entPercnt;
  }
  elsif ($chIn eq "|") {
    $szEntity = $entVbar;
  }
  else {
    $szEntity = $chIn;
  }
  return $szEntity;
} # end of Ch2Entity (coding char to SGML entity name)
 
sub IsWS($) # checks input string for whitespace; treturns 1 if it is ####
{
  my ($szIn) = @_;
  my $iStart;
  my $ch;

  for ($iStart = 0; $iStart < length($szIn); $iStart++) {
    $ch = substr $szIn, $iStart, 1;
    if ("$ch" ne " " &&
        "$ch" ne "\t" &&
        "$ch" ne "\n" &&
        "$ch" ne "\r" &&     
        "$ch" ne "\f") {
      return 0;
    }
  } # end of string: no non-whitespace found
  return 1; # all whitespace
} # end of IsWS, whitespace ##############################################

sub StripWS($) # strip whitespace from beginning & end of string #########
{
  my ($szIn) = @_;
  my $iStart;
  my $iEnd;
  my $ch;
  my $szReturn;

  # find first non-whitespace from start of string:
  for ($iStart = 0; $iStart < length($szIn); $iStart++) {
    $ch = substr $szIn, $iStart, 1;
    if (!IsWS($ch)) { last; }
  }
  # find first non-whitespace from end but not earlier than $iStart:
  for ($iEnd = length($szIn) - 1; $iEnd >= $iStart; $iEnd--) {
    $ch = substr $szIn, $iEnd, 1;
    if (!IsWS($ch)) { last; }
  }
  if ($iEnd >= $iStart) { # something non-whitespace left there
    $szReturn = substr $szIn, $iStart, ($iEnd - $iStart + 1);
  }
  else { # empty string (nl and/or whitespace only)
    $szReturn = "";
  }
  return $szReturn;
} # end of StripWS #######################################################

sub StripWSandCode($) # strip whitespace, convert code to ISO if needed
{
  my ($szIn) = @_;
  my $szStripped;
  my $szCoded;

  $szStripped = StripWS($szIn);
  if ($fInputCode == $WINCP1250) {
    $szCoded = wee2il2($szStripped);
    return $szCoded;
  }
  else {
    return $szStripped;
  }
} # end of StripWSandCode ################################################

sub GetDictionary($) # open and read dictionary ##########################
{
  my ($szDictFN) = @_; # get parameters
  my $iRC; # return code from this sub
  my $szDict7;

  $iRC = 0;
  $cEndingCountMax = 0;
  $fStripAccents = 0;
  $DICTIONARY = "<" . $szDictFN;
  print STDERR $szPN, ": Opening Dictionary ", $szDictFN, "\n";
  $iRCopen = open(DICTIONARY,$szDictFN);
  if (!defined $iRCopen) { # check if open ok
    # not found; try without the '7' before suffix:
    ($szDict7 = $szDictFN) =~ s/7(.[^.]*)$/\1/;
    $DICTIONARY = "<" . $szDict7;
    print STDERR $szPN, ": Trying (Full) Dictionary ", $szDict7, "\n";
    $iRCopen = open(DICTIONARY,$szDict7);
    if (!defined $iRCopen) { # check if open ok
      return 1;
    }
    else {
      $fStripAccents = 1;
    }
  }
  # read dictionary one line at a time:
  $szDictIDAndVersionInfo = "[Version info not available]";
  while (<DICTIONARY>) {
    chomp; # get rid of final whitespace incl. newline
    if ("$_" eq "") { next; }
    $chID = substr $_, 0, 1;
    if ($chID eq ";") {
      if ($_ =~ /[ ]@[ ]/) {
        ($szDummy,$szDictIDAndVersionInfo) = split /@[ ]/, $_, 2;
      }
      # otherwise ordinary comment; ignore it
    }
    elsif ($chID eq "E") { # "ending" line
      ($chIgn,$szPat,$fNeg,$fNej,$szEnding,$szTagc,$szTagp,$szJunk) =
	  split /\|/, $_; # splitting the input line into:
                         # paradigm name ($szPat)
                         # negation flag (0/1: $fNeg)
                         # superlative flag (0/1: $fNej)
                         # ending string ($szEnding)
                         # compact tag ($szTagc), positional tag ($szTagp)
      if (defined $szJunk ||
          !defined $szPat ||
          !defined $fNeg ||
          !defined $fNej ||
          !defined $szEnding ||
          !defined $szTagc ||
          !defined $szTagp) { # check if all fields present and nothing else
        print STDERR $szPN,
                   ": ERROR: ending data error: ",$DictionaryFileName,
                   ": line ", $_, "\n";
        return 3;
      }

      if ($fStripAccents == 1) {
        $szEnding = StripAccents($szEnding);
      }
      # remember the longets ending's length:
      if (length($szEnding) > $cEndingCountMax) {
        $cEndingCountMax = length($szEnding);
      }
      # save the ending data (array of five items):
      @rgEndData = ($szPat, $fNeg, $fNej, $szTagc, $szTagp);
      push @{ $hEnding2PdgmEtc{$szEnding} }, [ @rgEndData ];
      $cEndingCount++;
    }    
    elsif ($chID eq "R") { # "root" line
      ($chIgn,$szPat,$szRoot,$szLemma,
              $szTagc,$szTagp,$chAltPos,$szSem,$szStyle,$szComment,$szJunk) =
	  split /\|/, $_; # splitting the input line into:
                         # paradigm name ($szPat)
                         # root string ($szRoot)
                         # lemma idemtifier ($szLemma)
                         # compact tag ($szTagc), positional tag ($szTagp)
                         # alternate POS character ($chAltPos), obsolete
                         # terminology / semantics ($szSem)
                         # style etc. ($szStyle)
                         # human-readable comment
      if (defined $szJunk ||
          !defined $szPat ||
          !defined $szRoot ||
          !defined $szLemma ||
          !defined $szSem ||
          !defined $szStyle ||
          !defined $szComment ||
          !defined $chAltPos ||
          !defined $szTagc ||
          !defined $szTagp) { # check if all fields present and nothing else
        print STDERR $szPN,
                   ": ERROR: root data error: ",$DictionaryFileName,
                   ": line ", $_, "\n";
        return 4;
      }
      if ($fStripAccents == 1) {
        $szRoot = StripAccents($szRoot);
      }
      # save the root data (array of seven items):
      @rgRootData = ($szLemma, $szTagc, $szTagp, $chAltPos, $szSem, $szStyle,
                     $szComment);
      push @{ $hRootPdgm2LemmaEtc{$szRoot}{$szPat} }, [ @rgRootData ];
      $cRootCount++;
    }    
    else { # wrong line ID: not a dictionary or a fomrat error
      print STDERR $szPN,
                   ": ERROR: dictionary format error: ",$DictionaryFileName,
                   ": line ", $_, "\n";
      return 2;
    }    
  } # end of dictionary reading loop: while (<DICTIONARY>)

  # print out the statistics:
  print STDERR $szPN,
               ": Dictionary ", $DictionaryFileName, " ",
               $szDictIDAndVersionInfo, "\n";
  print STDERR $szPN,
               ": Dictionary ", $DictionaryFileName, " read in; ",
               $cEndingCount, " ending, ", 
               $cRootCount, " root entries.\n";
  print STDERR $szPN, ": Max ending length: ", $cEndingCountMax, "\n";

  return 0;

} # end of GetDictionary subroutine ######################################

sub InitializeReader($) # input file & reading routine initalization ######
{
  my $iRC = 0;

  my ($szInputData) = @_; # nonempty if input in parameter, 
                          # otherwise read STDIN
  $szLookAheadLine = ""; # initialize lookahead line
  $szLookAhead = ""; # lookahead string (rest of current line)
  $szCurrentLine = StripWSandCode($szInputData); # current input line

  # get beginning of input: either file or parameter
  if ($szCurrentLine eq "") {
    while (<STDIN>) { # in fact, read once unless empty line(s) found at beg
      chomp;
      $szCurrentLine = StripWSandCode($_);
      #print STDERR "read raw: $_, $szCurrentLine\n"; # dbg
      if ("$szCurrentLine" ne "") { last; }
    }
    if ("$szCurrentLine" eq "") {
      return 1; # empty input file?
    }
  } # no data in parameter
  #print STDERR "read in: ", "|", $szCurrentLine, "|\n"; # dbg
  if ($szCurrentLine =~ /^[ \t\r\f\n]*<[hH][tT][mM][lL][ >]/) {
    $fInputType = $fSGML;
    $fSGMLType = $fHTML;
  }
  elsif (
     $szCurrentLine =~ /^[ \t\r\f\n]*<[cC][sS][tT][sS][ >]/ ||
     $szCurrentLine =~
 /^[ \t\r\f\n]*<[!][dD][oO][cC][tT][yY][pP][eE].*[^a-zA-Z0-9][cC][sS][tT][sS]/
      ) {
    $fInputType = $fCSTS;
    $fSpacingInfo = 1; # must be set otherwise the output is screwed up
  }
  elsif ($szCurrentLine =~ /^[ \t\r\f\n]*</) {
    $fInputType = $fSGML;
    $fSGMLType = $fSGML;
  }
  else {
    $fInputType = $fPlain;
  }
  #print STDERR $szPN, ": Input file type: ", 
  #             $fInputType, " ", $fSGMLType, "\n"; # dbg

  $szLookAhead = $szCurrentLine;
  return $iRC;
} # end of InitializeReader ##############################################

sub InitializeWriter() # output file & writing initalization #############
{
  my $iRC = 0;

  if ($szSGMLDTD ne "" && ($fInputType != $fCSTS)) {
    print "<$szSGMLDTD>\n"; # output heading: csts DTD used throughout
                            # if not changed in SetOptions
  }

  return $iRC;
} # end of InitializeWriter ##############################################

sub CloseReader() # input file & reading routine close ###################
{
  my $iRC = 0;

  return $iRC;
} # end of CloseReader ###################################################

sub CloseWriter() # output file & writing close ##########################
{
  my $iRC = 0;

  if ($szSGMLDTD ne "" && ($fInputType != $fCSTS)) {
    print "</$szSGMLDTD>\n"; # output heading close: csts DTD used throughout
                             # if not changed in SetOptions
  }

  return $iRC;
} # end of CloseWriter ###################################################

sub FormAttributes($) # analyze form for digits, capitalization etc. #####
{
                      # returns SGML attr value: "" (for lowercase)
                      # or mixed, upper, cap, num
  my $iRC = 0;
  my ($szFormIn) = @_;

  if ($szFormIn =~ /^[0-9]+([.,][0-9]+)*$/) {
    return "num";
  }
  if ($szFormIn =~ /[0-9]/) { # number somewhere
    return "mixed";
  }
  if ($szFormIn =~ /^[A-Z]$/ ||
      $szFormIn =~ /^($reUpperAcc)$/) { # single upper: cap
    return "cap";
  }
  if ($szFormIn =~ /^(([A-Z])|($reUpperAcc))+$/) { # all up > 2: upper
    return "upper";
  }
  if ($szFormIn =~ /^[A-Z]/ ||
      $szFormIn =~ /^($reUpperAcc)/) { # first uppercase, not
                                       # all: cap
    return "cap";
  }
  if ($szFormIn =~ /[A-Z]/ ||
      $szFormIn =~ /($reUpperAcc)/) { # upper in the middle
    return "mixed";
  }
  return ""; # default: all real letters, lowercase
} # end of FormAttributes ################################################

sub LetterToLower($) # lowercase one letter ##############################
{
  my ($ch) = @_;
  if ($ch =~ /^[A-Z]$/) { # ascii uppercase letter
    return (chr ((ord $ch) + 32)); 
  }
  if ((ord $ch) < 128) { # other ascii: return unchanged
    return $ch;
  }
  $i = index $szUpperAccented, $ch;
  if ($i < 0) { # not an uppercase accented letter: return unchanged
    return $ch;
  }
  else { # accented uppercase:
    return substr $szLowerAccented, $i, 1;
  }
} # of lowercase one letter ##############################################

sub LowercaseDown($) # lowercase all letters #############################
{
  my ($szIn) = @_;
  my $iLen; # length of input
  my $i; # current letter
  my $szResult;

  $szResult = "";
  $iLen = length $szIn;
  for ($i = 0; $i < $iLen; $i++) {
    $szResult .= LetterToLower(substr $szIn, $i, 1);
  }
  return $szResult;
} # end of LowercaseDown - lowercasing everything ########################

sub CapitalizeDown($) # lowercase all letters but the first ##############
{
  my ($szIn) = @_;
  my $szResult = substr $szIn, 0, 1; # keep first letter as is
  $szResult .= LowercaseDown(substr $szIn, 1); # lowercase the rest
  return $szResult; # return 'capitalized' version
} # end of CapitalizeDown - lowercasing everything but the first letter ##

sub GetToken() # The tokenizer: get one word form, plus surrounding junk #
{
  # returns the results in global variables:
  #    $szGapBef       junk before wordform, SGML-coded
  #    $szGapAft       junk after wordform, SGML-coded
  #    $szForm         wordform as found in input
  #    $szAfterAnalysis original tags in fCSTS markup to be preserved
  #    $szSpaceAfterForm tag to put on next line (NoSpace tag, usually <D>)
  #    $szSGMLTagForm  wordform tag (w/attributes if needed)
  #    $fDotFollows    period/fullstop immediately followed wordform in input
  #    $fHyphenFollows hyphen immediately followed wordform in input
  # returns 0 if eof

  my $iRC = 1; # still something on input (assuming...)
  my $fSGMLcontinues; # set to 1 if SGML continues over several lines

  # initialize Output variables:
  $szGapBef = "";
  $szGapAft = "";
  $szForm = "";
  $szSGMLTagForm = "";
  $szAfterAnalysis = "";
  $szSpaceAfterForm = "";
  $fDotFollows = 0;
  $fHyphenFollows = 0;

  $fSGMLContinues = 0;
  
# print STDERR "GetToken() entered, lookahead (20 max): ", 
#       (substr $szLookAhead, 0, 20), "\n"; # dbg

  SomethingToRead:
  while ($iRC == 1) { # still something to read
    while ("$szLookAhead" ne "") {
      $chFirst = substr $szLookAhead, 0, 1; # first character of input
      if ($fInputType == $fSGML 
          && "$chFirst" eq "&") { # input "character" is an entity name
	$ch = "&";
        $szLookAhead = substr $szLookAhead, 1, length($szLookAhead) - 1;
	while ($szLookAhead ne "" && "$ch" ne ";") {
          $ch = substr $szLookAhead, 0, 1; # first character of lookahead
          $szLookAhead = substr $szLookAhead, 1, length($szLookAhead) - 1;
          $chFirst .= $ch;
        }
      } # of reading an entity name (like e.g. &name;)
      # ----------------- csts code ----------------------------------------
      if (($chFirst eq "<" && $fInputType == $fCSTS)) { # SGML tag in csts;
        # first of all, replace spaces which creeped in instead of newlines
        # to newlines:
	$szLookAhead = StripWS($szLookAhead);
        $szLookAhead =~ s/[ \t\r\f]+[<]/\n</g;
        # get first "line" from lookahead:
        ($szActLine, $szLookAhead) = split /[\n]/, $szLookAhead, 2; # split
                                     # at first newline
#print STDERR "actline $szActLine, lookahead before split: |$szLookAhead|\n"; # dbg
        if ($szLookAhead eq "") { # just one tag on line
          # we need here at least one more line in szLookAhead, to check
          # for <D>; only if not running as server (everything should
          # be one line - maybe with spaces - if running as server
	  if ($fServ == 0) {
            $szStdinLine = <STDIN>;
            $szStdinLine = StripWSandCode($szStdinLine);
            if (defined($szStdinLine)) {
              ($szLookAhead = $szStdinLine) =~ s/[ \t]+[<]/\n</g;
              # $szLookAhead not empty if still not at end of STDIN
              if ($szLookAhead =~ /^<D>$/) {
                # no-space symbol; must read next line:
                $szStdinLine = <STDIN>;
                $szStdinLine = StripWSandCode($szStdinLine);
                if (defined($szStdinLine)) {
                  $szStdinLine =~ s/[ \t]+[<]/\n</g;
                  $szLookAhead .= "\n" . $szStdinLine; # append to lookahead
	        }
              }
            }
	  } # end of not server mode
          # else: in server mode, it's there or not from input line...
        }
        # from now on, deal only with $szActLine here (fCSTS)
        # (and check szLookAhead only for <D>)
#print STDERR "working on part: !$szActLine!\n"; # dbg
        # copy if not <d> or <f>:
	if ($szActLine =~ /^<[fd][ >]/) { # punctuation or form
          # ok, first analyze the line for pre-analyzed stuff:
          # and even before that, get original f/d tag and form:
          # don't forget, $szActLine does not contain \n here!
          ($szSGMLTagForm, $szActLineRest) = split />/, $szActLine, 2;
          $szSGMLTagForm .= ">"; # add lost >
          ($szForm,$szActLineRest) = split /</, $szActLineRest, 2;
          while ($szActLineRest ne "") {
            ($szActTagText,$szActLineRest) = split /</, $szActLineRest, 2;
            if ($szActTagText =~ /^MM[lt]>/) {
              # ignore incoming MMt/MMl tags since morphology will replace them
              # but mark the position in AfterAnalysis:
              if ($szAfterAnalysis =~ /<>/) {
                # already there because of some previously found <t> tag:
                # remove it, must go here (MMl/MMt has precedence)
		  $szAfterAnalysis =~ s/<>//g;
              }
              $szAfterAnalysis .= "<>";
            }
            if ($szActTagText =~ /^t>/ &&
                $szActLineRest !~ /^t>/) {
              # if [last] tag present in input, put it here, and put <>
              #  for morphology result after that
              $szAfterAnalysis .= "<" . $szActTagText . "<>";
            }
            else { # keep tag incl. leading < opening symbol
              $szAfterAnalysis .= "<" . $szActTagText;
            }
          } # end of something afer form/punct in ActLine
          # $szActLineRest empty here
          # find out if dot or hyphen follows after a no-space tag:
          if (($szLookAhead ne "")  &&
              ($szLookAhead =~ /^<D>/)) { # don't even ask about fSpacingInfo
            if ($szLookAhead =~/^<D>[ \f\r\t\n]*[\n][ \f\r\t\n]*<d>[.]/) {
               $fDotFollows =  1;
	    }
            elsif ($szLookAhead =~/^<D>[ \f\r\t\n]*[\n][ \f\r\t\n]*<d>[-]/) {
               $fHyphenFollows = 1;
	    }
          }
          if (($szAfterAnalysis !~ "") &&
              ($szAfterAnalysis !~ /<>/)) {
	    $szAfterAnalysis = "<>" . $szAfterAnalysis; # if no original
                # morphology in input, put (eventually) result at the beginning
          }
          return $iRCTokenOK;
        }
        else { # other SGML - just copy to output
          chomp $szActLine;
          # print STDERR "just copying: !$szActLine!\n"; # dbg
          $szGapBef .= $szActLine . "\n";
	  $szActLine = ""; # $szLookAhead already set
        }
      } # end of csts code 
      # ----------------- SGML code ----------------------------------------
      elsif (($chFirst eq "<" && $fInputType == $fSGML) ||
           $fSGMLContinues != 0) { # SGML tag;
        # convert to "<i>"-type <csts> tag - even if it continues on next line:
        $szGapBef .= $szSGMLJunk;
        $i = 0;
        $fSGMLContinues = 1; # default; for next iteration if '>' not found
        SGMLread:
        while ($i < length($szLookAhead)) {
          $chFirst = substr $szLookAhead, $i, 1;
	  if ($chFirst eq ">") {
            $szGapBef .= $clEnt . "\n";
            $i++;
            $fSGMLContinues = 0;
            last SGMLread;
	  }
	  else {
            $szT = Ch2Entity($chFirst);
            $szGapBef .= $szT;
  	  }
	}
        continue { $i++; }
        if ($i == length($szLookAhead)) {
	  $szLookAhead = ""; # nothing else left on line
          chomp $szGapBef; # eliminate extra nl (if tag ends at end of line)
          $szGapBef .= "\n";
	}
        else { # shift lookahead over extracted SGML tag:
          $szLookAhead = substr $szLookAhead, $i, length($szLookAhead) - $i;
	}
        if ($fIgnoreJunk == 1) { # no output of <i>-tags requested:
          $szGapBef = ""; # delete it...
        }
      }
      # ------------------- wordforms (and numbers) --------------------------
      elsif ($chFirst =~ /^[a-zA-Z0-9]$/ ||
          $chFirst =~ /^($reLowerAcc)$/o ||
          $chFirst =~ /^($reUpperAcc)$/o) { # start of word form
                                                # or number
        while ("$szLookAhead" ne "") {
          $szForm .= $chFirst; # add character to token output string
          $szLookAhead = substr $szLookAhead, 1, length($szLookAhead) - 1;
                             # remove character from lookahead          
          $chFirst = substr $szLookAhead, 0, 1; # first character of lookahead
          if (!(($chFirst =~ /^[a-zA-Z0-9]$/) ||
                ($chFirst =~ /^($reLowerAcc)$/) ||
                ($chFirst =~ /^($reUpperAcc)$/))) { # word form
            if (($szForm =~ /^[0-9]+$/) && # all digits so far
                (($chFirst eq '.') || ($chFirst eq ',')) # dec. point in CZ
                && (((substr $szLookAhead, 1, 1) . "X")
                                          =~ /^[0-9]X$/)) # digit must follow
            {
              # ok, it's a decimal point
            }
            else {
              last;
            }
	  }
        } # end of while adding letters & digits
        # decide on dot/hyphen flags: 
        if ("$szLookAhead" ne "") { # i.e., not at end of line
          if ("$chFirst" eq ".") { $fDotFollows = 1; } 
          elsif ("$chFirst" eq "-") { $fHyphenFollows = 1; } 
	}
        # else at end of line, i.e. not dot, no hyphen immediately follows
        $szSGMLTagForm = "<$szSGMLForm";
        $szAttrCap = FormAttributes($szForm);
        if ("$szAttrCap" eq "") {
	  $szSGMLTagForm .= ">";
	}
        else {
	  $szSGMLTagForm .= " " . $szAttrCap . ">";
        }
        if ($szLookAhead ne ""  &&
            (!(IsWS($chFirst))) &&
            ($fSpacingInfo == 1)) {
          $szSpaceAfterForm = $szSGMLNoSpace;
        }
        return $iRCTokenOK;
      }
      # ----------------- whitespace (ignored) -----------------------------
     elsif (IsWS($chFirst)) { # Whitespace
        $szLookAhead = StripWS($szLookAhead); # get rid of whitespace at beg
        # do not return here; go back to tokenize $szLookAhead or read new
      }
      # ----------------- nbsp (nonbreakable space) treated as <i> ---------
      elsif ($chFirst eq "&nbsp;") { # nonbreakable whitespace
        $szGapBef .= $szSGMLJunk . $chFirst . "\n";
        # do not return here; go back to tokenize $szLookAhead or read new
        if ($fIgnoreJunk == 1) { # no output of <i>-tags requested:
          $szGapBef = ""; # oh well, delete it...
        }
      }
      # ---------------- delimiters ------------------------------------------
      elsif ($chFirst =~ 
          /^[-<>!^.,;:\/?\$@#~`%&*()_+={}\]\[|\\\"\'\200-\377]$/)
        { # delimiter (punctuation)
        $szT = Ch2Entity($chFirst);
        # obsolete: $szGapBef .= "<d>" . $szT . "\n";
        $szSGMLTagForm = $szSGMLPunctuation;
        $szForm = $szT;
        # move lookahead just one character:
        $szLookAhead = substr $szLookAhead, 1, length($szLookAhead) - 1;
        $chFirst = substr $szLookAhead, 0, 1; # first character of lookahead
        # check if space after the current delimiter:
        if ($szLookAhead ne ""  &&
            (!(IsWS($chFirst))) &&
            ($fSpacingInfo == 1)) {
          $szSpaceAfterForm = $szSGMLNoSpace;
        }
        return $iRCTokenOK;
      }
      # ---------------- else: just junk ------------------------------------
      else { # convert junk to <i>-type tags too (like SGML tags)
        $szT = sprintf "&#0x%02x;", ord $chFirst;
        $szGapBef .= $szSGMLJunk . $szT . "\n";
        $szLookAhead = substr $szLookAhead, 1, length($szLookAhead) - 1;
        if ($fIgnoreJunk == 1) { # no output of <i>-tags requested:
          $szGapBef = ""; # oh well, delete it...
        }
      }
    } # while lookahead not empty
    $szCurrentLine = "";
    if ($fServ == 1) { # server mode: do not read STDIN
      return 0; # lookahead empty, nothing more to do
    }
    while (<STDIN>) { # in fact, just read once unless empty line(s) found
      chomp;
      $szCurrentLine = StripWSandCode($_);
      if ("$szCurrentLine" eq "") { next; }
      # print STDERR "read in: ", $szCurrentLine, "\n"; # dbg
      last;
    } # end of reading stdin (while (<>))
    $szLookAhead = $szCurrentLine;
    if ("$szLookAhead" eq "") { 
      $iRC = 0; }
  } # of while reading (until form found)
  if ($szLookAhead ne ""  &&
      (!(IsWS($chFirst))) &&
      ($fSpacingInfo == 1)) {
    $szSpaceAfterForm = $szSGMLNoSpace;
  }
  return $iRC;
} # end of tokenizer GetToken (getting wordform from input) ################

sub FilterTag($$) ###### filters tag out (based on SetOptions) #############
{
  my $iResult = 0; # default: do not filter out
  my ($szTagCurr, $szSuffList) = @_;
  # check if output filtering requested:
  if ($szSuffList ne "") { # check if tag should be
                                                # filtered out (per option)
    if ($fPositionalTags) { # check last position
      $chTag = substr $szTagCurr, length($szTagCurr) - 1, 1; # last position
      if (index($szSuffList,$chTag) >= 0) { # found, del
        $iResult = 1; # ignore this result completely
      }
    } # end of positional tag check
    else { # checking compact tag:
      $iTagLen = length $szTagCurr;
      if ($iTagLen >= 2) { # tag suffix -<sym> possibly present
        if (substr $szTagCurr, $iTagLen - 2, 1 eq "-") { # yes...
          $chTag = substr $szTagCurr, $iTagLen - 1, 1; # last ch
          if (index($szSuffList, $chTag) >= 0) { # found
            $iResult = 1; # ignore this result completely
	  }
	} # end of tag suffix found in filter list
      } # end of tag long enough to check
    } # end of compact tag check for filtering out
  } # end of option to filter out based on tag set
  return $iResult;
} # end of FilterTag #######################################################

sub FilterString($$) ##### signals if string1 intersect with filter string #
{
  # signals if result should be ignored (rc == 1) based on presence of
  # any char from string1 in string 2;
  # ie. returns 1 if intersection non empty

  my $iResult = 0; # default: do not filter out
  my ($szStrCurr, $szSuffList) = @_; # $szSuffList is a string of chars
                                     # leading to filtering out of a result
                                     # if found in current string ($szStrCurr)
  my $i = 0; # for index
  my $cLen = 0; # length of option string containing letters to be filtered
  my $ch; # character from current string to be tested

  # check if output filtering requested:
  if ($szSuffList ne "" &&
      $szStrCurr ne "") { # ...and if anything to test in actual data
    $cLen = length($szStrCurr);
    for ($i = 0; $i < $cLen; $i++) {
      $ch = substr($szStrCurr, $i, 1);
      if (index($szSuffList,$ch) >= 0) { # found: filter it out
	$iResult = 1;
	last;
      } # end of found branch: set return result to "FILTER OUT" and exit loop
    } # end of loop over actual characters in current string
  } # end of if: filtering requested & makes sense
  return $iResult;
} # end of FilterString: signalling if filtering should occur ##############

sub Template2Tag($$$)
{
  my ($szTagTemplate, $fNegP, $fNejP) = @_;
  my $i;
  my $chTemplate;
  my $chTag;
  my $szTagResult = "";
  for ($i = 0; $i < length($szTagTemplate); $i++) {
    $chTemplate = substr $szTagTemplate, $i, 1;
    if ($chTemplate eq "@") { # negation; replace by A/N depending on presence
                              # of negative prefix in original form
      if ($fNegP == 1) { $chTag = "N"; }
      else { $chTag = "A"; }
    }
    elsif ($chTemplate eq "#") { # superlative template char; replace by 2/3
                                 # (comparative/superlative) depending on
                                 # the presence of the superlative prefix
      if ($fNejP == 1) { $chTag = "3"; }
      else { $chTag = "2"; }
    }
    else { # regular non-template char
      $chTag = $chTemplate;
    }
    $szTagResult .= $chTag; # add to resulting tag being built
  }
  return $szTagResult;  
} # end of Template2Tag

sub MorphAnalyzeBareForm($$$$$) # Analyze one wordform; case fixed, no pref #
{

  my $iRC = 0;
  # get parameters:
  my ($szFormIn, $fDot, $fHyph, $fNegStripped, $fNejStripped) = @_;

  my $iEnding; # current ending length
  my @rgRootDataCurr;
  my @rgrgEndCurr; # list of all data records for a current ending
  my @rgEndDataCurr;

#  print STDERR " - Analyzing: $szFormIn, flags dot $fDot hyph $fHyph,",
#                      " neg $fNegStripped, nej $fNejStripped \n"; # dbg

  $szFormInNP = $szFormIn;
  for ($iEnding = 0; $iEnding <= $cEndingCountMax; $iEnding++) {
    #
    # main analysis loop; successively segmenting input form into
    # root - ending pair, then checking for a paradigm match; if successful,
    # all morphological information is collected from both ending and root
    # and stored in $hResultsCollected (hash)
    #
    # start from the end, i.e. with root == form and ending == 0
    #
    if ($iEnding >= length($szFormInNP)) { last; } # finish if no root left
    if ($iEnding == 0) { # just started:
      $szEndingCurr = "0"; # for formal match with ending data
      $szRootCurr = $szFormInNP; # set root to full form
    }
    else { # second and successive passes go through here:
      $szEndingCurr = substr $szFormInNP, length($szFormInNP) - $iEnding, 
                                          $iEnding; # ending construction
      $szRootCurr = substr $szFormInNP, 0, length($szFormInNP) - $iEnding;
                    # root construction
    }
    # root + ending set now in $sz...Curr
    if (defined $hRootPdgm2LemmaEtc{$szRootCurr}) { # just for efficiency! 
      foreach $rgEndCurr (@{ $hEnding2PdgmEtc{$szEndingCurr} }) {
        # for every data entry in the current ending list,
        # find out if the current root matches:
        ($szPatCurr, $fNegCurr, $fNejCurr, $szTagcCurr, $szTagpCurr) = 
          @{ $rgEndCurr }; 
        # first filter out wrong analyses because of "external" conditions,
        # i.e. prefix mismatch, period/hyphen followup mismatch:
        if ($fNegCurr >= $fNegStripped) { # negation allowed or neg. prefix
                                          # not present
	  if ($fNejCurr >= $fNejStripped) { # sup. prefix allowed or
                                            # not present
            if ($szPatCurr eq "zkr" && $fDot == 0) { next; } # period req'd 
                                                             # but not present
            if ($szPatCurr eq "poml" && $fHyph == 0) { next; } # hyph req'd
            # ok, now really get the root data:
            if (exists $hRootPdgm2LemmaEtc{$szRootCurr}{$szPatCurr}) { 
              # match found!
              foreach $rgRootCurr 
                (@{$hRootPdgm2LemmaEtc{$szRootCurr}{$szPatCurr}}) {
                # for all roots which also match paradigm, get root info:
		($szLemmaDict, $szTagcDict, $szTagpDict, $chAltPosDict, 
                  $szSemDict, $szStyleDict, $szCommentDict) =
                  @{ $rgRootCurr};
                # first, build the tag part of result: ($szSGMLTagOut):
                if ($fPositionalTags) {
    	          if ($szTagpDict ne "0") {
                    $szTagCurrTemplate = $szTagpDict;
		  }
		  else {
                    $szTagCurrTemplate = $szTagpCurr;
		  }
		} # positional tag ouput request selected by user
		else { # compact tag output
    	          if ($szTagcDict ne "0") {
                    $szTagCurrTemplate = $szTagcDict;
		  }
		  else {
                    $szTagCurrTemplate = $szTagcCurr;
		  }
		} # compact tag output selected by user
                # replace variables (@,#) for neg/superlative:
                $szTagCurr = Template2Tag($szTagCurrTemplate,
                                          $fNegStripped,
                                          $fNejStripped);
                # tag ready now
                if (FilterTag($szTagCurr,$szFilterT) > 0) {
                  next; # ignore current result completely
		}
                # prepare additional info (sem/style):
                $szAddInfo = "";
                if ($chAltPosDict ne "0") {
		  if (FilterString($chAltPosDict,$szFilterL) > 0) { next; }
	          for ($i = 0; $i < length($chAltPosDict); $i++) {
                    $chAdd = substr $chAltPosDict, $i, 1;
                    if ($chAdd ne substr $szTagCurr, 0, 1) {
                      $szAddInfo .= "_:" . $chAdd; # add if different than
                                                   # tag POS (comp/pos same!)
		    }
		  }
                } # end of adding syntax (alternate POS) info
                if ($szSemDict ne "0") {
		  if (FilterString($szSemDict,$szFilterL) > 0) { next; }
	          for ($i = 0; $i < length($szSemDict); $i++) {
                    $szAddInfo .= "_;" . substr $szSemDict, $i, 1;
		  }
                } # end of adding 'semantics' (terminology) info
                if ($szStyleDict ne "0") {
		  if (FilterString($szStyleDict,$szFilterL) > 0) { next; }
	          for ($i = 0; $i < length($szStyleDict); $i++) {
                    $szAddInfo .= "_," . substr $szStyleDict, $i, 1;
		  }
                } # end of adding style info
                if ($fCommentsOut == 1) {
                  if ($szCommentDict ne "0") {
                   $szAddInfo .= "_^"; # comment start separator 
	            for ($i = 0; $i < length($szCommentDict); $i++) {
                      $chAdd = substr $szCommentDict, $i, 1;
                      $szAddInfo .= Ch2Entity($chAdd); # eliminates potential 
                                                       # SGML problems
		    } # end of converting to entities where needed
	          } # end of comment contains something
                } # end of comment output
                $hResultsCollected{$szLemmaDict}{$szTagCurr} = $szAddInfo;
                                             # save result; replace prev.
                                             # if [pure] lemma & tag identical
	      } # end of collecting all dict data after a match had been ok'd
            } # end of root data exists for root/pdgm, i.e. match found
          } # end of superlative prefix match
        } # end of negative prefix match
      } # end of for each ending data record 
    } # root exists in dictionary (speedup test only)
    # else it makes no sense to check ending data - cannot be analyzed anyway

  } # of for (segmenting form to root + ending)

  return $iRC;
} # end of MorphAnalyzeBareForm ############################################

sub MorphAnalyzePrefixedForm($$$) # Analyze one wordform; case fixed #######
                                  # solve prefixing (ne-, nej-)
{
  my $iRC = 0;
  # get parameters:
  my ($szFormIn, $fDot, $fHyph) = @_;
  my $fNegPresent = 0;
  my $fNejPresent = 0;
  my $szFormWithoutPrefix;

  # first, call morph analysis with unchanged form:
  $iRCCall = MorphAnalyzeBareForm($szFormIn,$fDot,$fHyph,0,0); 
  if ($iRCCall != 0) { return 1; }
  if ($szFormIn =~ /^[nN][eE]/) { # form starts with ne-
    $szFormWithoutPrefix = substr $szFormIn, 2;
    $iRCCall = MorphAnalyzeBareForm($szFormWithoutPrefix,$fDot,$fHyph,1,0); 
    if ($iRCCall != 0) { return 2; }
  }
  if ($szFormIn =~ /^[nN][eE][jJ]/) { # form starts with nej-
    $szFormWithoutPrefix = substr $szFormIn, 3;
    $iRCCall = MorphAnalyzeBareForm($szFormWithoutPrefix,$fDot,$fHyph,0,1); 
    if ($iRCCall != 0) { return 3; }
  }
  if ($szFormIn =~ /^[nN][eE][jJ][nN][eE]/) { # form starts with nejne-
    $szFormWithoutPrefix = substr $szFormIn, 5;
    $iRCCall = MorphAnalyzeBareForm($szFormWithoutPrefix,$fDot,$fHyph,1,1); 
    if ($iRCCall != 0) { return 4; }
  }
  return 0;
} # end of MorphAnalyzePrefixedForm ########################################

sub MorphAnalyzeCapitalizedForm($$$$) # Analyze form; solve UC if any ######
        # solve capitalization:
        # - if original form in all uppercase or mixed: 
        #                 call analysis 3 times: UC/mix --> Cap --> LC
        # - if original form Capitalized only (or mixed w/no UC elsewhere):
        #                 call analysis twice: Cap --> LC
        # - if original lowercased:
        #                 LC only, one call only
        # Further (Cap/LC) analysis called regardless whether 
        #          success or not previously. (This rule might change
        #          unless the tagger takes care of it (preferred!))
{
  my $iRC = 0;
  # get parameters:
  my ($szFormIn, $szSGMLTag, $fDot, $fHyph) = @_;
  my $iCapitalizationLevel = 0; # 0 for unchanged, 1 for Capitalized only,
                                # 2 for all lowercase
  my $szFormCapOnly;
  my $szFormLower;

  # set global check/filter strings based on pass (1/2):
  if ($iPass == $STRICT) {
    $szFilterT = $szTagSuffixIgnore . $szTagSuffixIgnoreIfOtherOK;
    $szFilterL = $szLexIgnore . $szLexIgnoreIfOtherOK;
  }
  elsif ($iPass == $RELAXED) {
    $szFilterT = $szTagSuffixIgnore;
    $szFilterL = $szLexIgnore;
  }
  else { # internal fatal error - wrong $iPass number
    return 99;
  } # end of setting filtering strings

  undef %hResultsCollected; # initialize hash for result collection

  for ($iCapitalizationLevel = $iCapOrig; # loop three times: original caps,
       $iCapitalizationLevel <= $iCapLower; # then capitalized (1st letter)
       $iCapitalizationLevel++) {           # then all lowercase
    if ($iCapitalizationLevel == 0) {
      # always do this level; input form unchanged
      $iRCCall = MorphAnalyzePrefixedForm($szFormIn,$fDot,$fHyph);
                    # result: appended to in $szAnalyzedToken
      if ($iRCCall != 0) {
	return 1;
      }
    } # end of first pass code
    elsif ($iCapitalizationLevel == 1) { # try capitalized (if different)
      $szFormCapOnly = CapitalizeDown($szFormIn);
      if ($szFormCapOnly eq $szFormIn) { # same as original form, do not proc.
        next;
      }
      $iRCCall = MorphAnalyzePrefixedForm($szFormCapOnly,$fDot,$fHyph);
                    # result: appended to in $szAnalyzedToken
      if ($iRCCall != 0) {
	return 2;
      }
    } # end of second pass code
    else { # all lowercase (if different from either of previous two)
      $szFormLower = LowercaseDown($szFormIn);
      if ($szFormLower eq $szFormIn ||
          $szFormLower eq $szCapOnly) { # already processed
        next;
      }
      $iRCCall = MorphAnalyzePrefixedForm($szFormLower,$fDot,$fHyph);
                    # result: appended to in $szAnalyzedToken
      if ($iRCCall != 0) {
	return 3;
      }
    } # end of third pass code
  } # of for loop for the capitalization levels
  return 0;
} # end of MorphAnalyzeCapitalizedForm #####################################

sub MorphAnalyzeForm($$$$) # Analyze one input form; solve capitalization ##
{
  my $iRC = 0;
  my $iRCCalled = 0;
  # get parameters:
  my ($szFormIn, $szSGMLTag, $fDot, $fHyph) = @_;

  undef %hResultsCollected; # initialize hash for result collection

  $iPass = $STRICT; # first pass: filter out everything incl. conditionals
                   #  -- see SetOptions
  $iRCCalled = MorphAnalyzeCapitalizedForm($szFormIn,$szSGMLTag,$fDot,$fHyph);
  if ($iRCCalled != 0) { # first pass failed fatally
    return 1;
  }
  if (%hResultsCollected <= 0) { # no result found after heavy 
                                 # (STRICT) filtering
    # try second pass, filter out only results unconditionally requested

    $iPass = $RELAXED;
    $iRCCalled = 
      MorphAnalyzeCapitalizedForm($szFormIn,$szSGMLTag,$fDot,$fHyph);
    if ($iRCCalled != 0) { # second pass failed fatally
      return 2;
    }
  }
  # output results now regardless if any success at any pass:
  # creating printable string of results:
  $szAnalyzedToken = ""; # prepare for concatenation of results
  if (%hResultsCollected > 0) { # yes, some analysis found (after filtering)
    foreach $szL (sort keys %hResultsCollected) { # for each lemma
      $szLemmaInfoOut = ""; # final lemma info (_....) to go out with $szL
      $szAnalyzedTokenTag = ""; # accumulated tags for current lemma in $szL
      $szAnalyzedToken .= $szSGMLLemmaOut . $szL; # Lemma output; 
                                                  # without info yet
      foreach $szT (sort keys %{ $hResultsCollected{$szL} }) {
        $szAnalyzedTokenTag .= $szSGMLTagOut . $szT; # add tag to list (string)
        $szLemmaInfo = $hResultsCollected{$szL}{$szT}; # get lemma info
        if (length($szLemmaInfo) > length($szLemmaInfoOut)) {
          if ($szLemmaInfoOut ne "") { # lemma info already set 
            print STDERR $szPN, ": incompatible lemma info ($szL/$szT)",
                                ": $szLemmaInfo vs. $szLemmaInfoOut", "\n";
	  } # end of printing error message
                # NB: this is kind of rough, if szLemmaInfoOut already set,
                # but probably works best; all
                # info should be same anyway for current lemma so we are only
                # making a soft-landing here for dictionary problems
          $szLemmaInfoOut = $szLemmaInfo; # replace old str (current is longer)
	} # end of storing lemma info anew
      } # end of loop over tags
      $szAnalyzedToken .= $szLemmaInfoOut . $szAnalyzedTokenTag;
    } # end of loop over lemmas
  } # end of: some result generated
  else { # "NOT FOUND": no solution found whatsoever (in the dictionary):
    if ($szSGMLTag eq $szSGMLPunctuation) { # "unknown" punctuation;
                                   # almost always handled here
      $szAnalyzedToken = $szSGMLLemmaOut . $szFormIn . $szSGMLTagOut;
      if ($fPositionalTags == 1) {
        $szAnalyzedToken .= "Z:-------------";
      }
      else { # compact tags requested 
        $szAnalyzedToken .= "ZIP";
      }
    } # end of handling: punctuation if not found
    elsif ($szSGMLTag =~ /^<($szSGMLForm)[ ]+(.*[ ])?num/) { 
                                              # number/digital expression
                                              # if not found in dict
      $szAnalyzedToken = $szSGMLLemmaOut . $szFormIn . $szSGMLTagOut;
      if ($fPositionalTags == 1) {
        $szAnalyzedToken .= "C=-------------";
      }
      else { # compact tags requested 
        $szAnalyzedToken .= "ZNUM";
      }
    } # end of handling: numerical/digital expressions
    else {
      $szAnalyzedToken = $szSGMLLemmaOut . $szFormIn . $szSGMLTagOut;
      if ($fPositionalTags == 1) {
        $szAnalyzedToken .= "X@-------------";
      }
      else { # compact tags requested 
        if (length($szFormIn) == 1) {
	  $szAnalyzedToken .= "NOMORPH1"; }
        else {
	  $szAnalyzedToken .= "NOMORPH"; }
      }
    } # end of handling: not a number, not punctuation
  } # end of final "NOT FOUND" handling section
  return 0;
} # end of MorphAnalyzeForm ################################################

sub MorphAnalyzeInput($) # morphologically analyze input (incl. tokenization)
{
  my $iRC = 0; # return code

  my ($szDataIn) = @_; # if nonempty, input in parameter, not STDIN

  $iRCCall = InitializeReader($szDataIn); # determine input file type,
                                          # set flags
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Input initialization failed; check input file (stdin empty?)\n";
    return 1;
  }
  $iRCCall = InitializeWriter(); # write initial SGML to output 
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Output initialization failed; check output file (stdout)\n";
    return 2;
  }

  ### main analyzer loop here: ###
  $fSomethingToRead = GetToken(); # read stuff into
                                  # $szGapBef, $szForm, $szGapAft,
                                  # $szSGMLTagForm,
                                  # $fDotFollows, $fHyphenFollows
  while ($fSomethingToRead != 0) {
              # print STDERR "GetToken returned:\n", "GB: ", $szGapBef, "\n", 
              #            "FM: ", $szForm, "\n", 
              #            "TG: ", $szSGMLTagForm, "\n", 
              #            "GA: ", $szGapAft, "\n",
              #            "DH: ", $fDotFollows,$fHyphenFollows,"\n"; # dbg
    $iRCCall = MorphAnalyzeForm($szForm,$szSGMLTagForm,
                                $fDotFollows,$fHyphenFollows);
                    # result: in $szAnalyzedToken, incl. $szForm and
                    # proper SGML tags
    if ($iRCCall != 0) {
      print STDERR $szPN,
                   ": ERROR analyzing ", $szForm, "; input line ", $iLineInput,
                   "; exiting\n";
      print "<<<ERROR>>>\n";
      $iRC = 6;
      last; # exit the loop;
    }
    else { # everything went ok
      if ("$szGapBef" ne "") {
	print $szGapBef; # newline already in there
      }
      if ($fNoTokenSGML == 0) { # suppresion of SGML tags in from of 
                                # tokens not req'd in SetOptions
        PrintCoded($szSGMLTagForm);
      }
      PrintCoded($szForm);
      $szAfterAnalysisBeforeResult = "";
      $szAfterAnalysisAfterResult = "";
      chomp $szAfterAnalysis;
      chomp $szAfterAnalysis;
      if ($szAfterAnalysis ne "") { # split to keep original order:
        ($szAfterAnalysisBeforeResult,$szAfterAnalysisAfterResult) 
          = split /<>/, $szAfterAnalysis, 2;
        $szAfterAnalysisAfterResult =~ s/<>//g; # get rid of remaining <>s 
      }
      if ($szAfterAnalysisBeforeResult ne "") {
        PrintCoded($szAfterAnalysisBeforeResult);
      }
      PrintCoded($szAnalyzedToken);
      if ($szAfterAnalysisAfterResult ne "") {
        PrintCoded($szAfterAnalysisAfterResult);
      }
      print "\n";
      if ($szSpaceAfterForm ne "") { # also empty if fSpacingInfo disabled
        PrintCoded($szSpaceAfterForm);
        print "\n";
      }
      if ("$szGapAft" ne "") {
	PrintCoded($szGapAft); # newline already in there
      }
    }

    $fSomethingToRead = GetToken(); # read next word; if everything ok, 
                                   # read stuff is in: $szGapBef, 
                                   # $szForm, $szSGMLTagForm, $szGapAft,
                                   # $fDotFollows, $fHyphenFollows
  } # end of the main word-reading loop (through GetToken())

  # still have to check for any gap-like material at the end of file:
  if ("$szGapBef" ne "") {
    PrintCoded($szGapBef); # newline already in there
  }
  if ("$szGapAft" ne "") {
    PrintCoded($szGapAft); # newline already in there
  }

  $iRCCall = CloseWriter(); # close output file (write final SGML tag(s))
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Output file close failed; check output file (stdout)\n";
    return 3;
  }
  $iRCCall = CloseReader(); # close input file
  if ($iRCCall != 0) {
    print STDERR $szPN,
      ": Input file close failed;\n";
    return 4;
  }

  return $iRC;
} # end of MorphAnalyzeInput #################################################

sub SetOptionsNews() ############## Setting options suitable for news ########
{
  $szLexIgnore = ""; # keep default (do not hard-ignore anything)
  $szLexIgnoreIfOtherOK = "xsanhelv"; # ignore if other solutions present
  $szTagSuffixIgnore = "";  # keep default (do not hard-delete based on tags)
  $szTagSuffixIgnoreIfOtherOK = "345679";  # ign if other present
} # end of SetOptionsNews ####################################################

sub SetOptionsColl() ############## Setting options suitable for coll style ##
{
  $szLexIgnore = ""; # keep default (do not hard-ignore anything)
  $szLexIgnoreIfOtherOK = "sanelv"; # ignore if other solutions present
  $szTagSuffixIgnore = "";  # keep default (do not hard-delete based on tags)
  $szTagSuffixIgnoreIfOtherOK = "3459";  # ign if other present
} # end of SetOptionsColl ####################################################

sub SetOptionsOld() ############### Setting options suitable for old czech ###
{
  $szLexIgnore = ""; # keep default (do not hard-ignore anything)
  $szLexIgnoreIfOtherOK = "elv"; # ignore if other solutions present
  $szTagSuffixIgnore = "";  # keep default (do not hard-delete based on tags)
  $szTagSuffixIgnoreIfOtherOK = "9";  # ign if other present
} # end of SetOptionsOld  ####################################################

sub SetOptionsSpell() ############# Delete almost everything ("spelling") ####
{
  $szLexIgnore = "xsanhelvtSHL"; # delete "strange" words (inc. surnames, etc.)
  $szLexIgnoreIfOtherOK = ""; # ignore if other solutions present
  $szTagSuffixIgnore = "3456789";  # selete many "strange tags"
  $szTagSuffixIgnoreIfOtherOK = "2";  # ign if other present
} # end of SetOptionsSpell ###################################################

sub SetOptionsAll() ############# Do not delete anything (full res.) #########
{
  $szLexIgnore = ""; # default: output everything regardless of lex
                     # (synt/sem/style) info
                     # Default: ""
  $szLexIgnoreIfOtherOK = ""; # default: output everything regardless of lex
                     # (synt/sem/style) info
                     # Default: ""
  $szTagSuffixIgnore = "";  # default: output everything 
                            # regardless of 
                            # tag suffix (-<dig/sym> in compact system,
                            # position 15 in positional)
                            # Default: ""
  $szTagSuffixIgnoreIfOtherOK = "";  # default: output everything 
                            # regardless of 
                            # tag suffix (-<dig/sym> in compact system, 
                            # position 15 in positional), if something
                            # other still left (without suffix/other suff.) 
                            # Default: ""
} # end of SetOptionsAll #####################################################

sub SetOptionsDemo() ############# Set demo (interactive) mode ###############
{
  $szSGMLDTD = ""; # no heading
  $szSGMLLemmaOut = "\n\t"; # lemma on next line, indented
  $szSGMLTagOut = "\n\t\t"; # POS-tags on next line, indented twice
  $fIgnoreJunk = 1; # 1: ignore SGML code and other junk on input
  $fSpacingInfo = 0; # no "no-spacing" info (<D>)
  $fNoTokenSGML = 1; # 1: do not output SGML tag in front of wordforms
  $fCommentsOut = 1; # 1: output the human-readable comments (stuff in paren)
  # keep SGML tags intact (irrelevant anyway)
} # end of SetOptionsDemo ####################################################

sub SetOptionsBatch() ############# Set batch mode (basic) ###################
{
  $szSGMLDTD = "csts"; # SGML tag name for output format DTD
                       # printed out as the first & last line
                       # !! Change NOT recommended.
                       # default: "csts"
                       # use "" for no DTD start/end tags at all
                       # NB: must change if you change any of the following
                       # SGML tags!
  $szSGMLLemmaOut = "<MMl>"; # lemma tag after morphological analysis
                             # Default: "<MMl>"
                             # Changing to non-SGML STRONGLY discouraged
                             # (only space or ctrl chars or weird
                             # characters > 128 which are not letters
                             # would work!!)
  $szSGMLTagOut = "<MMt>"; # POS-tags tag after morphological analysis
                           # Default: "<MMt>"
                           # Changing to non-SGML STRONGLY discouraged
                           # (only space or ctrl chars or weird
                           # characters > 128 which are not letters
                           # would work!!)
  $fIgnoreJunk = 0; # 1: ignore SGML code and other junk on input
                    # default: 0: output everything (SGML-safely coded)
                    # in <i> tags, or whatever $szSGMLJunk says
  $fSpacingInfo = 1; # 1: provide spacing info by including a NO-SPACE tag
                     #    "<D>" (or whatever $szSGMLNoSpace is) between
                     #    any two tokens originally without space between
                     #    them (such as in "... then they went home." -->
                     #    <f>home \n <D> \n <d>.)
                     # Default: 1
  $fNoTokenSGML = 0; # 1: do not output SGML tag in front of wordforms
                     #    nor punctuation (normally, <f[ attr]> and <d>).
                     # Default: 0
                     # !! Change to 1 NOT recommended.
                     #    NB: overwrites change in $szSGMLForm and 
                     #        $szSGMLPunctuation
                     #    Useful when no sgml output required (do not
                     #      forget to disable also Junk, SpacingInfo, DTD
                     #      and change $szSGMLLemmaOut and $szSGMLTagOut
                     #      if you really want SGMLless output).
  $szSGMLForm = "f"; # Output SGML tag name for wordform (incl. numbers
                       # in digits)
                       # USed also internally: must not be empty string!
                       # Default: "f"
                       # Change strongly discouraged (to disable: see
                       #             $fNoTokenSGML)
  $szSGMLPunctuation = "<d>"; # Output SGML tag for punctuation
                              # USed also internally: must not be empty string!
                              # Default: "<d>"
                              # Change to non-SGML discouraged (to disable: see
                              #             $fNoTokenSGML)
  $szSGMLNoSpace = "<D>"; # tag to be inserted between two tokens withOUT
                          # a whitespace originally (if $fSpacingInfo set)
                          # Change to non-SGML discouraged (to disable: see
                          #             $fSpacingInfo)
  $szSGMLJunk = "<i>"; # SGML output tag for preserving input SGML codes
                       # and other junk (makes sense only if $fIgnoreJunk == 1)
                       # Change to non-SGML discouraged (to disable: see
                       #             $fIgnoreJunk)
  $fPositionalTags = 1; # 1: output in positional tag system
                        # 0: compact
                        # Default: 1 (positional tags)
  return 0;
} # end of SetOptionsBatch ###################################################

sub SetOptionSet($) ##### Set Options based on parameter (string) ############
{
  my $iRC = 0;
  my ($szOptionSet) = @_;
  if ($szOptionSet eq "news") {
    SetOptionsNews();
  }
  elsif ($szOptionSet eq "coll") {
    SetOptionsColl();
  }
  elsif ($szOptionSet eq "old") {
    SetOptionsOld();
  }
  elsif ($szOptionSet eq "spell") {
    SetOptionsSpell();
  }
  elsif ($szOptionSet eq "all") {
    SetOptionsAll();
  }
  elsif ($szOptionSet eq "demo") {
    SetOptionsDemo();
  }
  elsif ($szOptionSet eq "batch") {
    SetOptionsBatch();
  }
  else {
    print STDERR "$szPN: Wrong filter parameter: $szOptionSet\n";
    $iRC = 1;
  }
  return $iRC;
} # end of SetOptionSet ######################################################

sub SetOptionsDefaults() ########### Setting runtime options (def) ###########
{
  # all options except for dictionary name should be set here
  # if need is out there for several different versions, make a copy
  # of this whole thing, rename and change options there...
  # see  also the individual overwrites (sub SetOptions...)
  my $iRC = 0;
  $iRC = SetOptionsBatch();
  if ($iRC != 0) { return $iRC; }
  SetOptionsAll();
  $fCommentsOut = 1; # 1: output the human-readable comments (stuff
                     # in parenthesis; some include derivational info)
                     # Default: 1
  return 0;
} # end of SetOptionsDefaults ################################################

############################################################## MAIN ##########

  # Program starts here:
  $iRC = 0;
  $szPN = $0; # remember program name for messages

  # check which code to expect at input: (might be changed dynamically):
  if ($szPN =~ /iso/i) {
    $fInputCode = $ISOLATIN2;
  }
  elsif ($szPN =~ /win/i) {
    $fInputCode = $WINCP1250;
  }
  else {
    $fInputCode = $ISOLATIN2; # default for old scripts
  }

  if ($fInputCode == $WINCP1250) {
    printf STDERR 
           "$szPN: using input/output code conversion from/to Win CP1250.\n";
  }

  # check if we should run as daemon (first parameter, shift rest if yes):
  $fServ = 0;
  if (($#ARGV >= 0) &&
      ($ARGV[0] eq $szSERVER)) {
    $fServ = 0;
    shift;
    if ($#ARGV >= 0) {
      $iPort = $ARGV[0];
    }
    else {
      print STDERR "Usage: SERVER port dict [mode] [mode]\n";
      exit 985;
    }
    # untaint port:
    if ($iPort =~ /^([0-9]+)$/) {
      $iPort = $1;
    }
    else {
      print STDERR "$szPN: $iPort must be numeric.\n";
      exit 986;
    }
    if ($iPort < 30523 &&
        $iPort > 30800) {
      print STDERR "$szPN: port $iPort out of range.\n";
      exit 988;
    }
    shift; # get ready for dictionary in
  } # end of daemon check

  if ($fServ == 0) { # not as server/daemon
    # check if we should use supplied filenames instead of stdin/out:
    if (($#ARGV >= 0) &&
        ($ARGV[0] eq $szNOSTDIO)) {
      shift;
      if (($#ARGV >= 0)) {
        $szMAIn = $ARGV[0];
        shift;
        if (($#ARGV >= 0)) {
	    $szMAOut = $ARGV[0];
            shift;
        } 
        else {
          print STDERR 
            "Usage: $szPN NOSTDIO Infile Outfile [usual params]...\n";
          exit 998;
        }
      }
      else {
        print STDERR 
          "Usage: $szPN NOSTDIO Infile Outfile [usual params]...\n";
        exit 997;
      }
      close(STDIN);
      open (STDIN, $szMAIn) || do {
        print STDERR "$szPN: cannot open input file $szMAIn\n";
        exit 996;
      };
      close(STDOUT);
      $iRCCall = open STDOUT, ">$szMAOut";
      if ($iRCCall == undef) { 
        print STDERR "$szPN: cannot open output file $szMAOut\n";
        exit 995;
      };
    } # end of NOSTDIO check
  } # end of not daemon

  # log version, calling style:
  print STDERR 
      "$szPN: version $FMVERSION $FMDATE";
  if ($fServ == 1) {
    print STDERR " SERVER";
  }
  else {
    if ($szMAIn ne "" ||
        $szMAOut ne "") {
      print STDERR " NOSTDIO";
      if ($szMAIn ne "") {
        print STDERR " In: $szMAIn";
      }
      if ($szMAOut ne "") {
        print STDERR " Out: $szMAOut";
      }
    }
    else {
      print STDERR " stdin/stdout";
    }
  } # logging; not server mode
  print STDERR "\n";

  # dictionary and mode parameters:
  if (($#ARGV < 0) || ($#ARGV > 2)) {
    print STDERR 
      "Usage: $szPN Dictionary-file [Mode] [Mode] < In > Out\n";
    print STDERR 
      "(or:   $szPN NOSTDIO Filein Fileout Dictionary-file [Mode] [Mode])\n";
    print STDERR
      "          Filtering Modes:\n";
    print STDERR
      "                              news      suitable for news\n";
    print STDERR
      "                              coll      colloquial style exp.\n";
    print STDERR
      "                              old       old Czech\n";
    print STDERR
      "                              spell     heavy filtering\n";
    print STDERR
      "                              all       no filtering (default)\n";
    print STDERR
      "          Demo mode:\n";
    print STDERR
      "                              demo      for interactive testing\n";
    exit 999;
  }
  $DictionaryFileName = $ARGV[0]; # get dictionary name #####################

  # remember optional command line arguments:
  if ($#ARGV > 0) {
    $szPar1 = $ARGV[1];
  }
  if ($#ARGV > 1) {
    $szPar2 = $ARGV[2];
  }

  # Initialize options: #####################################################
  $iRC = SetOptionsDefaults();
  if ($iRC != 0) {
    print STDERR $szPN,
                 ": ERROR: cannot set options ",
                 "; exiting\n";
    exit 1;
  }
  # set additional options according to parameter 2:
  if ($szPar1 ne "") {
    $iRC = SetOptionSet($szPar1);
    if ($iRC != 0) {
      print STDERR $szPN,
                   ": ERROR: cannot set additional options (1)",
                   "; exiting\n";
      exit 2;
    }
  } # end of setting additional parameters
  # set additional options according to parameter 3:
  if ($szPar2 ne "") {
    $iRC = SetOptionSet($szPar2);
    if ($iRC != 0) {
      print STDERR $szPN,
                   ": ERROR: cannot set additional options (2)",
                   "; exiting\n";
      exit 3;
    }
  } # end of setting additional parameters

  $iRC = GetDictionary($DictionaryFileName); # open and read dictionary in ###
  if ($iRC != 0) { # exit if any problems encountered
    print STDERR $szPN,
                 ": ERROR: cannot open and/or read dictionary ",
                 $DictionaryFileName, "; exiting\n";
    exit (10 + $iRC);
  }


  if ($fServ == 0) { # not server mode 
    $iRC = MorphAnalyzeInput(""); ###########################################
    if ($iRC != 0) { # exit if any problems encountered
      print STDERR $szPN,
              ": ERROR: cannot morphologically analyze input file (STDIN)",
              "; exiting\n";
      exit (100+$iRC);
    }
  }
  else { # server code follows
    sub REAPER {
      $waitedpid = wait;
      # loathe sysV: it makes us not only reinstate
      # the handler, but place it after the wait
      $SIG{CHLD} = \&REAPER;
    }
    sub logmsg { print STDERR "$szPN @_ at ", scalar localtime, "\n" }
    $szDataInCmd = "0f3b"; # command code for server: data follows
    $szFromWhCmd = "8b01"; # command code for server: client IP
    $szSetOptCmd = "a10a"; # command code for server: option follows
    $szStopCmd = "02ab"; # command code for server: stop
    $szVerCmd = "e7ce"; # command code for client: dict version info
    $szEofCmd = "800d"; # command code for client: end of data
    BEGIN { $ENV{PATH} = '/usr/ucb:/bin' }
    $EOL = "\012";

    my $iProto = getprotobyname('tcp');

    socket(Server, PF_INET, SOCK_STREAM, $iProto) 
      || die "socket: $!";
    setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
      || die "setsockopt: $!";
    bind(Server, sockaddr_in($iPort, INADDR_ANY))
      || die "bind: $!";
    listen(Server,1) 
      || die "listen: $!";

    logmsg "server started on port $iPort, 1 conn.";

    my $iPaddr;

    $SIG{CHLD} = \&REAPER;

    # main client-listening loop:
    for ( ; $iPaddr = accept(Client,Server); close Client) {
      my ($iPort,$iAddr) = sockaddr_in($iPaddr);

#      my $szClientHostname = gethostbyaddr($iAddr,AF_INET);
#      logmsg "connection from $szClientHostname [",
#                   inet_ntoa($iAddr), "]",
#                   "at port $iPort"; # comment out once it runs well
#                                     # (hopefully only from us)

      open(STDIN,"<&Client") or die "can't dup STDIN";
      open(STDOUT,">&Client") or die "can't dup STDOUT";
      STDOUT->autoflush();
      $szIn = <STDIN>;
      chomp $szIn;
      $fDataExpected = 0;
      $szClient = "unknown";
      $szLastOpt = "default";
      if (defined($szIn)) {
        $fDataExpected = 1;
        while ($szIn !~ /^$szDataInCmd[ ]/o) {
          $cLengthIn = length($szIn);
          if ($cLengthIn > 400) {
            logmsg "Too long input ($cLengthIn). Aborting.";
	    exit 912;
          }
          if ($szIn =~ /^$szSetOptCmd[ ]/o) {
            $szLastOpt = substr($szIn,5);
	    $iRC = SetOptionSet($szLastOpt);
	    if ($iRC != 0) {
	      logmsg "invalid option $szIn";
  	      $fDataExpected = 0;
	      last;
	    }
          }
          elsif ($szIn =~ /^$szFromWhCmd[ ]/o) {
            $szClient = substr($szIn,5);
          }
          elsif ($szIn =~ /^$szStopCmd[ ]/o) {
	      logmsg "Stop command received. Exitting.";
	    exit 990;
          }
          else { # wrong command code!
            logmsg "Unauthorized command code. Hacker's attack?";
	    exit 911;
          }
          $szIn = <STDIN>;
          if (!defined($szIn)) {
	    $fDataExpected = 0;
	    last;
          }
          chomp $szIn;
        }
      } 
      if ($fDataExpected == 1) { # yes, work on it
        if ($szClient eq "unknown") {
          logmsg "Client address unknown. Cannot serve uknown clients, you hacker.";
          sleep 5;
          exit 915;
	}
        $cLengthIn = length($szIn);
        if ($cLengthIn > 400) {
          logmsg "Too long input ($cLengthIn). Aborting.";
          exit 913;
        }
        $szDataToAnalyze = substr($szIn,5);
        logmsg "$szDataToAnalyze (F: $szLastOpt) for $szClient";
        print STDOUT "$szVerCmd ", "Code version: ", $FMVERSION, " ", $FMDATE, 
                                   "@ Dictionary: ",
                                   "$szDictIDAndVersionInfo ",
                                   "@ ",
                                   "Endings: $cEndingCount, ",
                                   "Roots: $cRootCount",
                                   $EOL;
        $iRC = MorphAnalyzeInput($szDataToAnalyze); ###########################
        if ($iRC != 0) { # exit if any problems encountered
          print STDERR $szPN,
           ": ERROR: cannot morphologically analyze input $szDataToAnalyze\n",
        }
        print STDOUT "$szEofCmd ", $EOL;
      }
      else { # no data (just set option, maybe)
        print STDOUT "$szEofCmd ", "ERROR", $EOL;
      }
      sleep 4; 
    } # end of server listening code loop
  } # end of server code

  0;
######################################################### END OF MAIN ########
