#!/usr/bin/perl
#
#
# Code file from any code to any code supported by HB_CharCode8
#    (goes through IsoL2 if needed source/dest not IsoL2)
#
# usage: 
#   for any-to-any code:  HBCode.pl in-code out-code < in-file > out-file
#
# usage alternatives for coding from/to ISO Latin2:
#   for coding from some defined code to ISO Latin2: 
#                         HBCode.pl from in-code < in-file > out-file
#   for coding to some defined code from ISO Latin2: 
#                         HBCode.pl to out-code < in-file > out-file
#
#

use HB_CharCode8;
use strict;

my $VERSION="1.0.0"; # initial version

my $iRC;
my $szPN = $0;
my @AvailableCodes;

if (@ARGV != 2) {
  print STDERR "Usage: $szPN in-code out-code < in-file > out-file\n";
  print STDERR "       $szPN from in-code < in-file > out-file\n";
  print STDERR "       $szPN to out-code < in-file > out-file\n";
  @AvailableCodes = HB_CharCode8ListCodes();
  print STDERR "\n";
  print STDERR "  Available code names for in-code/out-code:\n\n";
  print STDERR "       @AvailableCodes\n\n";
  exit 1;
}

my $szIncode = lc($ARGV[0]);
my $szOutcode = lc($ARGV[1]);

my $szInter; # intermediate coded line (ISO Latin 2)
my $szOut; # output coded line

if ($szIncode eq "from") {
  $szIncode = $szOutcode;
  $szOutcode = "iso-8859-2";
}
elsif ($szIncode eq "to") {
  $szIncode = "iso-8859-2";
}

$iRC = HB_CharCode8SetTable($szIncode);
if ($iRC != 0) {
  print STDERR "$szPN: cannot initialize input table $szIncode (sp?)\n";
  exit 6;
}
$iRC = HB_CharCode8SetTable($szOutcode);
if ($iRC != 0) {
  print STDERR "$szPN: cannot initialize output table $szOutcode (sp?)\n";
  exit 7;
}
if ($szIncode eq $szOutcode) {
  print STDERR "$szPN: Warning: I am not sure what you are trying to do, but ",
               "going through ISO L2 here might not be lossless.\n";
}

# don't bother with double coding if in/out code is ISO Latin2:

if ($szIncode eq "iso-8859-2") {
  $iRC = HB_CharCode8SetTable($szOutcode);
  if ($iRC != 0) {
    print STDERR "$szPN: cannot initialize table $szOutcode (sp?)\n";
    exit 2;
  }
  while (<STDIN>) {
    chomp;
    $szOut = HB_CharCode8ToString($_);
    print "$szOut\n";
  }
}
elsif ($szOutcode eq "iso-8859-2") {
  $iRC = HB_CharCode8SetTable($szIncode);
  if ($iRC != 0) {
    print STDERR "$szPN: cannot initialize table $szIncode (sp?)\n";
    exit 3;
  }
  while (<STDIN>) {
    chomp;
    $szOut = HB_CharCode8FromString($_);
    print "$szOut\n";
  }
}
else {
  while (<STDIN>) {
    chomp;
    $iRC = HB_CharCode8SetTable($szIncode);
    if ($iRC != 0) {
      print STDERR "$szPN: cannot switch to table $szIncode\n";
      exit 4;
    }
    $szInter = HB_CharCode8FromString($_);
    $iRC = HB_CharCode8SetTable($szOutcode);
    if ($iRC != 0) {
      print STDERR "$szPN: cannot switch to table $szOutcode\n";
      exit 5;
    }
    $szOut = HB_CharCode8ToString($szInter);
    print "$szOut\n";
  }
}

0;
